"""Home to various lifetime models, for use in dynamic stock modelling."""

from abc import abstractmethod
import numpy as np
import scipy.stats
from pydantic import BaseModel as PydanticBaseModel, model_validator
from typing import Any

# from scipy.special import gammaln, logsumexp
# from scipy.optimize import root_scalar

from .dimensions import DimensionSet, Dimension
from .flodym_arrays import FlodymArray
from .gauss_lobatto import gl_nodes, gl_weights


class UnevenTimeDim(PydanticBaseModel):

    dim: Dimension
    _bounds: np.ndarray = None

    @property
    def bounds(self):
        if self._bounds is None:
            self.compute_t_bounds()
        return self._bounds

    @property
    def interval_lengths(self):
        """Returns the length of the time intervals, i.e. the difference between the bounds."""
        return np.diff(self.bounds)

    def compute_t_bounds(self):
        middle = (np.array(self.dim.items[:-1]) + np.array(self.dim.items[1:])) / 2.0
        self._bounds = np.concatenate(
            (
                [middle[0] - (middle[1] - middle[0])],
                middle,
                [middle[-1] + (middle[-1] - middle[-2])],
            )
        )


class LifetimeModel(PydanticBaseModel):
    """Contains shared functionality across the various lifetime models."""

    dims: DimensionSet
    time_letter: str = "t"
    inflow_at: str = "middle"
    """If no quadrature is used, all inflow happens at one point in time, either at the beginning
    of the time period (start), in the middle (middle) or at the end (end).
    """
    n_pts_per_interval: int = 1
    """Inflow into the stock is in reality quite uniform over time, while survival factors only
    take into account a single point in time. This can be alleviated here by using numerical integration
    over each inflow time period with n points. A value of 1 means that the inflow
    is only evaluated once, depending on the inflow_at parameter. This is the default and should be used
    for long life times, e.g. >> 1 year.
    For n_pts > 1, the inflow is evaluated at n points in time within the time period.
    inflow_at is ignored in this case.
    A value of 2 means evaluation at the beginning and end of the time period, while higher
    values additionally evaluate at more points within the time period.
    Higher point numbers are only needed for short life times, e.g. < 1 year.
    Default is 1, meaning that the inflow is evaluated only once per time period.
    """
    _sf: np.ndarray = None
    _pdf: np.ndarray = None
    _t: UnevenTimeDim = None

    @model_validator(mode="after")
    def check_inflow_at(self):
        if self.inflow_at not in ["start", "middle", "end"]:
            raise ValueError("inflow_at must be one of 'start', 'middle', or 'end'.")
        return self

    @model_validator(mode="after")
    def cast_prms(self):
        for prm_name, prm in self.prms.items():
            if prm is not None:
                setattr(self, prm_name, self.cast_any_to_np_array(prm))
        return self

    @model_validator(mode="after")
    def init_t(self):
        if self.dims.letters[0] != self.time_letter:
            raise ValueError(
                f"Lifetime model expects time dimension to be the first dimension. "
                f"Lifetime model uses time_letter '{self.time_letter}', but first dimension is {self.dims.letters[0]}. "
                f"If the time dimension uses a different letter, set time_letter explicitly."
            )
        self._t = UnevenTimeDim(dim=self.dims[self.time_letter])
        return self

    @property
    @abstractmethod
    def prms(self) -> dict[str, np.ndarray | None]:
        raise NotImplementedError

    def _check_prms_set(self):
        for prm_name, prm in self.prms.items():
            if prm is None:
                raise ValueError(f"Lifetime {prm_name} must be set before use.")

    @property
    def shape(self):
        return self.dims.shape

    @property
    def _n_t(self):
        return self._t.dim.len

    @property
    def _shape_cohort(self):
        return (self._n_t,) + self.shape

    @property
    def _shape_no_t(self):
        return tuple(list(self.shape)[1:])

    @property
    def sf(self):
        if self._sf is None:
            self._sf = np.zeros(self._shape_cohort)
            self.compute_survival_factor()
        return self._sf

    @property
    def pdf(self):
        if self._pdf is None:
            self._pdf = np.zeros(self._shape_cohort)
            self.compute_outflow_pdf()
        return self._pdf

    def _tile(self, a: np.ndarray) -> np.ndarray:
        """tiles the input array a to the shape of the lifetime model, by adding non-time dimensions

        Args:
            a (np.ndarray): either of shape (n_t,) or (n_t, n_t), where the second dimension
                corresponds to the age cohort

        Returns:
            np.ndarray: the tiled array
        """
        index = (slice(None),) * a.ndim + (np.newaxis,) * len(self._shape_no_t)
        out = a[index]
        return np.tile(out, self._shape_no_t)

    def _remaining_ages(self, m, eta):
        t = eta * self._t.bounds[m + 1] + (1 - eta) * self._t.bounds[m]
        return self._tile(self._t.bounds[m + 1 :] - t)

    def compute_survival_factor(self):
        """Survival table self.sf(m,n) denotes the share of an inflow in year n (age-cohort) still
        present at the end of year m (after m-n years).
        The computation is self.sf(m,n) = ProbDist.sf(m-n), where ProbDist is the appropriate
        scipy function for the lifetime model chosen.
        For lifetimes 0 the sf is also 0, meaning that the age-cohort leaves during the same year
        of the inflow.
        The method compute outflow_sf returns an array year-by-cohort of the surviving fraction of
        a flow added to stock in year m (aka cohort m) in in year n. This value equals sf(n,m).
        This is the only method for the inflow-driven model where the lifetime distribution directly
        enters the computation.
        All other stock variables are determined by mass balance.
        The shape of the output sf array is NoofYears * NoofYears,
        and the meaning is years by age-cohorts.
        The method does nothing if the sf alreay exists.
        For example, sf could be assigned to the dynamic stock model from an exogenous computation
        to save time.
        """
        self._check_prms_set()
        quad_eta, quad_weights = self.get_quad_points_and_weights()
        for m in range(0, self._n_t):  # cohort index
            for eta, weight in zip(list(quad_eta), list(quad_weights)):
                t = self._remaining_ages(m, eta)
                self._sf[m::, m, ...] += weight * self._survival_by_year_id(t, m)

    def get_quad_points_and_weights(self):
        """Returns the quadrature points and weights for the inflow time periods."""
        if self.n_pts_per_interval > 10:
            raise ValueError("quad_order must be between 0 and 9.")
        if self.n_pts_per_interval > 1:
            nodes = [(x + 1) / 2 for x in gl_nodes[self.n_pts_per_interval]]
            weights = [w / 2 for w in gl_weights[self.n_pts_per_interval]]
            return nodes, weights
        else:
            if self.inflow_at == "start":
                return [0], [1]
            elif self.inflow_at == "middle":
                return [0.5], [1]
            elif self.inflow_at == "end":
                return [1], [1]

    @abstractmethod
    def _survival_by_year_id(m, **kwargs):
        pass

    @abstractmethod
    def set_prms(self):
        pass

    def cast_any_to_np_array(self, prm_in):
        if isinstance(prm_in, FlodymArray):
            prm_out = prm_in.cast_to(target_dims=self.dims).values
        else:
            prm_out = np.ndarray(self.shape)
            prm_out[...] = prm_in
        return prm_out

    def compute_outflow_pdf(self):
        """Returns an array year-by-cohort of the probability that an item
        added to stock in year m (aka cohort m) leaves in in year n. This value equals pdf(n,m).
        """
        t_diag_indices = np.diag_indices(self._n_t) + (slice(None),) * len(self._shape_no_t)
        self._pdf[t_diag_indices] = 1.0 - np.moveaxis(self.sf.diagonal(0, 0, 1), -1, 0)
        for m in range(0, self._n_t):
            self._pdf[m + 1 :, m, ...] = -1 * np.diff(self.sf[m:, m, ...], axis=0)


class FixedLifetime(LifetimeModel):
    """Fixed lifetime, age-cohort leaves the stock in the model year when a certain age,
    specified as 'Mean', is reached."""

    mean: Any = None

    @property
    def prms(self):
        return {"mean": self.mean}

    def set_prms(self, mean: FlodymArray):
        self.mean = self.cast_any_to_np_array(mean)

    def _survival_by_year_id(self, t, m):
        # Example: if lt is 3.5 years fixed, product will still be there after 0, 1, 2, and 3 years,
        # gone after 4 years.
        return (t < self.mean[m, ...]).astype(int)


class StandardDeviationLifetimeModel(LifetimeModel):
    mean: Any = None
    std: Any = None

    @property
    def prms(self):
        return {"mean": self.mean, "std": self.std}

    def set_prms(self, mean: FlodymArray, std: FlodymArray):
        self.mean = self.cast_any_to_np_array(mean)
        self.std = self.cast_any_to_np_array(std)


class NormalLifetime(StandardDeviationLifetimeModel):
    """Normally distributed lifetime with mean and standard deviation.
    Watch out for nonzero values, for negative ages, no correction or truncation done here.
    NOTE: As normal distributions have nonzero pdf for negative ages,
    which are physically impossible, these outflow contributions can either be ignored (
    violates the mass balance) or allocated to the zeroth year of residence,
    the latter being implemented in the method compute compute_o_c_from_s_c.
    As alternative, use lognormal or folded normal distribution options.
    """

    def _survival_by_year_id(self, t, m):
        if np.min(self.mean) < 0:
            raise ValueError("mean must be greater than zero.")

        return scipy.stats.norm.sf(
            t,
            loc=self.mean[m, ...],
            scale=self.std[m, ...],
        )


class FoldedNormalLifetime(StandardDeviationLifetimeModel):
    """Folded normal distribution, cf. https://en.wikipedia.org/wiki/Folded_normal_distribution
    NOTE: call this with the parameters of the normal distribution mu and sigma of curve
    BEFORE folding, curve after folding will have different mu and sigma.
    """

    def _survival_by_year_id(self, t, m):
        if np.min(self.mean) < 0:
            raise ValueError("mean must be greater than zero.")

        return scipy.stats.foldnorm.sf(
            t,
            self.mean[m, ...] / self.std[m, ...],
            0,
            scale=self.std[m, ...],
        )


class LogNormalLifetime(StandardDeviationLifetimeModel):
    """Lognormal distribution
    Here, the mean and stddev of the lognormal curve, not those of the underlying normal
    distribution, need to be specified!
    Values chosen according to description on
    https://docs.scipy.org/doc/scipy-0.13.0/reference/generated/scipy.stats.lognorm.html
    Same result as EXCEL function "=LOGNORM.VERT(x;LT_LN;SG_LN;TRUE)"
    """

    def _survival_by_year_id(self, t, m):
        mean_square = self.mean[m, ...] * self.mean[m, ...]
        std_square = self.std[m, ...] * self.std[m, ...]
        new_mean = np.log(mean_square / np.sqrt(mean_square + std_square))
        new_std = np.sqrt(np.log(1 + std_square / mean_square))
        # compute survival function
        sf_m = scipy.stats.lognorm.sf(t, s=new_std, loc=0, scale=np.exp(new_mean))
        return sf_m


class WeibullLifetime(LifetimeModel):
    """Weibull distribution with standard definition of scale and shape parameters."""

    weibull_shape: Any = None
    weibull_scale: Any = None

    @property
    def prms(self):
        return {"weibull_shape": self.weibull_shape, "weibull_scale": self.weibull_scale}

    def set_prms(self, weibull_shape: FlodymArray, weibull_scale: FlodymArray):
        self.weibull_shape = self.cast_any_to_np_array(weibull_shape)
        self.weibull_scale = self.cast_any_to_np_array(weibull_scale)

    def _survival_by_year_id(self, t, m):
        if np.min(self.weibull_shape) < 0:
            raise ValueError("Lifetime weibull_shape must be positive for Weibull distribution.")

        return scipy.stats.weibull_min.sf(
            t,
            c=self.weibull_shape[m, ...],
            loc=0,
            scale=self.weibull_scale[m, ...],
        )

    # @staticmethod
    # def weibull_c_scale_from_mean_std(mean, std):
    #     """Compute Weibull parameters c and scale from mean and standard deviation.
    #     Works on scalars.
    #     Taken from https://github.com/scipy/scipy/issues/12134#issuecomment-1214031574.
    #     """
    #     def r(c, mean, std):
    #         log_mean, log_std = np.log(mean), np.log(std)
    #         # np.pi*1j is the log of -1
    #         logratio = (logsumexp([gammaln(1 + 2/c) - 2*gammaln(1+1/c), np.pi*1j])
    #                     - 2*log_std + 2*log_mean)
    #         return np.real(logratio)

    #     # Maybe a bit simpler; doesn't seem to be substantially different numerically
    #     # def r(c, mean, std):
    #     #     logratio = (gammaln(1 + 2/c) - 2*gammaln(1+1/c) -
    #     #                 logsumexp([2*log_std - 2*log_mean, 0]))
    #     #     return logratio

    #     # other methods are more efficient, but I've seen TOMS748 return garbage
    #     res = root_scalar(r, args=(mean, std), method='bisect',
    #                     bracket=[1e-300, 1e300], maxiter=2000, xtol=1e-16)
    #     assert res.converged
    #     c = res.root
    #     scale = np.exp(np.log(mean) - gammaln(1 + 1/c))
    #     return c, scale



"""Home to various `Stock` classes,
including flow-driven stocks and dynamic (lifetime-based) stocks.
"""

from abc import abstractmethod
import numpy as np
from scipy.linalg import solve_triangular
from pydantic import BaseModel as PydanticBaseModel, ConfigDict, model_validator
from typing import Optional, Union
import logging

from .processes import Process
from .flodym_arrays import StockArray, FlodymArray
from .dimensions import DimensionSet
from .lifetime_models import LifetimeModel, UnevenTimeDim


class Stock(PydanticBaseModel):
    """Stock objects are components of an MFASystem, where materials can accumulate over time.
    They consist of three :py:class:`flodym.FlodymArray` objects:
    stock (the accumulation), inflow, outflow.

    The base class only allows to compute the stock from known inflow and outflow.
    The subclasses allows computations using a lifetime distribution function,
    which is necessary if not both inflow and outflow are known.
    """

    model_config = ConfigDict(protected_namespaces=(), arbitrary_types_allowed=True)

    dims: DimensionSet
    """Dimensions of the stock, inflow, and outflow arrays. Time must be the first dimension."""
    stock: Optional[StockArray] = None
    """Accumulation of the stock"""
    inflow: Optional[StockArray] = None
    """Inflow into the stock"""
    outflow: Optional[StockArray] = None
    """Outflow from the stock"""
    name: Optional[str] = "unnamed"
    """Name of the stock"""
    process: Optional[Process] = None
    """Process the stock is associated with, if any. Needed for example for the mass balance."""
    time_letter: str = "t"
    """Letter of the time dimension in the dimensions set, to make sure it's the first one."""
    _t: UnevenTimeDim = None

    @model_validator(mode="after")
    def validate_stock_arrays(self):
        if self.stock is None:
            self.stock = StockArray(dims=self.dims, name=f"{self.name}_stock")
        elif self.stock.dims.letters != self.dims.letters:
            raise ValueError(
                f"Stock dimensions {self.stock.dims.letters} do not match prescribed dims {self.dims.letters}."
            )
        if self.inflow is None:
            self.inflow = StockArray(dims=self.dims, name=f"{self.name}_inflow")
        elif self.inflow.dims.letters != self.dims.letters:
            raise ValueError(
                f"Inflow dimensions {self.inflow.dims.letters} do not match prescribed dims {self.dims.letters}."
            )
        if self.outflow is None:
            self.outflow = StockArray(dims=self.dims, name=f"{self.name}_outflow")
        elif self.outflow.dims.letters != self.dims.letters:
            raise ValueError(
                f"Outflow dimensions {self.outflow.dims.letters} do not match prescribed dims {self.dims.letters}."
            )
        return self

    @model_validator(mode="after")
    def validate_time_first_dim(self):
        if self.dims.letters[0] != self.time_letter:
            raise ValueError(
                f"Time dimension must be the first dimension, i.e. time_letter (now {self.time_letter}) must be the first letter in dims.letters (now {self.dims.letters[0]})."
            )
        return self

    @model_validator(mode="after")
    def init_t(self):
        self._t = UnevenTimeDim(dim=self.dims[self.time_letter])
        return self

    @abstractmethod
    def compute(self):
        # always add this check first
        self._check_needed_arrays()

    @abstractmethod
    def _check_needed_arrays(self):
        pass

    @property
    def shape(self) -> tuple:
        """Shape of the stock, inflow, outflow arrays, defined by the dimensions."""
        return self.dims.shape

    @property
    def process_id(self) -> int:
        """ID of the process the stock is associated with."""
        return self.process.id

    def to_stock_type(self, desired_stock_type: type, **kwargs):
        """Return an object of a new stock type with values and dimensions the same as the original.
        `**kwargs` can be used to pass additional model attributes as required by the desired stock
        type, if these are not contained in the original stock type.
        """
        return desired_stock_type(**self.__dict__, **kwargs)

    def check_stock_balance(self):
        balance = self.get_stock_balance()
        balance = np.max(np.abs(balance).sum(axis=0))
        if balance > 1:  # 1 tonne accuracy
            raise RuntimeError("Stock balance for dynamic stock model is too high: " + str(balance))
        elif balance > 0.001:
            print("Stock balance for model dynamic stock model is noteworthy: " + str(balance))

    def get_stock_balance(self) -> np.ndarray:
        """Check whether inflow, outflow, and stock are balanced.
        If possible, the method returns the vector 'Balance', where Balance = inflow - outflow - stock_change
        """
        dsdt = np.diff(
            self.stock.values, axis=0, prepend=0
        )  # stock_change(t) = stock(t) - stock(t-1)
        return self.inflow.values - self.outflow.values - dsdt

    def _to_whole_period(self, annual_flow: np.ndarray) -> np.ndarray:
        """multiply annual flow by interval length to get flow over whole period."""
        return np.einsum("t...,t->t...", annual_flow, self._t.interval_lengths)

    def _to_annual(self, whole_period_flow: np.ndarray) -> np.ndarray:
        """divide flow over whole period by interval length to get annual flow"""
        return np.einsum("t...,t->t...", whole_period_flow, 1.0 / self._t.interval_lengths)

    def __str__(self):
        base = f"{self.__class__.__name__} '{self.name}'"
        dims = f" with dims ({','.join(self.dims.letters)}) and shape {self.shape};"
        return base + dims


class SimpleFlowDrivenStock(Stock):
    """Given inflows and outflows, the stock can be calculated without a lifetime model or cohorts."""

    def _check_needed_arrays(self):
        if (
            np.max(np.abs(self.inflow.values)) < 1e-10
            and np.max(np.abs(self.outflow.values)) < 1e-10
        ):
            logging.warning("Inflow and Outflow are zero. This will lead to a zero stock.")

    def compute(self):
        self._check_needed_arrays()
        annual_net_inflow = self.inflow.values - self.outflow.values
        net_inflow_whole_period = self._to_whole_period(annual_net_inflow)
        self.stock.values[...] = np.cumsum(net_inflow_whole_period, axis=0)


class DynamicStockModel(Stock):
    """Parent class for dynamic stock models, which are based on stocks having a specified
    lifetime (distribution).
    """

    lifetime_model: Union[LifetimeModel, type]
    """Lifetime model, which contains the lifetime distribution function.
    Can be input either as a LifetimeModel subclass, or as an instance of a
    LifetimeModel subclass. For available subclasses, see `flodym.lifetime_models`.
    """
    _outflow_by_cohort: np.ndarray = None
    _stock_by_cohort: np.ndarray = None

    @model_validator(mode="after")
    def init_cohort_arrays(self):
        self._stock_by_cohort = np.zeros(self._shape_cohort)
        self._outflow_by_cohort = np.zeros(self._shape_cohort)
        return self

    @model_validator(mode="after")
    def init_lifetime_model(self):
        if isinstance(self.lifetime_model, type):
            if not issubclass(self.lifetime_model, LifetimeModel):
                raise ValueError("lifetime_model must be a subclass of LifetimeModel.")
            self.lifetime_model = self.lifetime_model(dims=self.dims, time_letter=self.time_letter)
        elif self.lifetime_model.dims.letters != self.dims.letters:
            raise ValueError("Lifetime model dimensions do not match stock dimensions.")
        return self

    def _check_needed_arrays(self):
        self.lifetime_model._check_prms_set()

    @property
    def _n_t(self) -> int:
        return list(self.shape)[0]

    @property
    def _shape_cohort(self) -> tuple:
        return (self._n_t,) + self.shape

    @property
    def _shape_no_t(self) -> tuple:
        return tuple(list(self.shape)[1:])

    @property
    def _t_diag_indices(self) -> tuple:
        return np.diag_indices(self._n_t) + (slice(None),) * len(self._shape_no_t)

    def get_outflow_by_cohort(self) -> np.ndarray:
        """Outflow by cohort, i.e. the outflow of each production year at each time step."""
        return self._outflow_by_cohort

    def get_stock_by_cohort(self) -> np.ndarray:
        """Stock by cohort, i.e. the stock of each production year at each time step."""
        return self._stock_by_cohort

    def _compute_outflow(self):
        self._outflow_by_cohort = np.einsum(
            "c...,tc...->tc...", self.inflow.values, self.lifetime_model.pdf
        )
        self.outflow.values[...] = self._outflow_by_cohort.sum(axis=1)

    def __str__(self):
        base = super().__str__()
        lifetime_model = self.lifetime_model.__class__.__name__
        return base + "\n  Lifetime model: " + lifetime_model


class InflowDrivenDSM(DynamicStockModel):
    """Inflow driven model.
    Given inflow and lifetime distribution calculate stocks and outflows.
    """

    def _check_needed_arrays(self):
        super()._check_needed_arrays()
        if np.allclose(self.inflow.values, np.zeros(self.shape)):
            logging.warning("Inflow is zero. This will lead to a zero stock and outflow.")

    def compute(self):
        """Determine stocks and outflows and store values in the class instance."""
        self._check_needed_arrays()
        self._compute_stock()
        self._compute_outflow()

    def _compute_stock(self):
        # for non-contiguous years, yearly inflow is multiplied with time interval length
        inflow_per_period = self._to_whole_period(self.inflow.values)
        self._stock_by_cohort = np.einsum(
            "c...,tc...->tc...", inflow_per_period, self.lifetime_model.sf
        )
        self.stock.values[...] = self._stock_by_cohort.sum(axis=1)


class StockDrivenDSM(DynamicStockModel):
    """Stock driven model.
    Given total stock and lifetime distribution, calculate inflows and outflows.
    This involves solving the lower triangular equation system A*x=b,
    where A is the survival function matrix, x is the inflow vector, and b is the stock vector.
    """

    solver: str = "manual"
    """Algorithm to use for solving the equation system.  Options are: "manual" (default), which uses
    an own python implementation, and "lapack", which calls the lapack trtrs routine via scipy.
    The lapack implementation may be more precise. Speed depends on the dimensionality,
    but the manual implementation is usually faster.
    """

    @model_validator(mode="after")
    def init_solver(self):
        if self.solver not in ["manual", "lapack"]:
            raise ValueError("Solver must be either 'manual' or 'lapack'.")
        return self

    def _check_needed_arrays(self):
        super()._check_needed_arrays()
        if np.allclose(self.stock.values, np.zeros(self.shape)):
            logging.warning("Stock is zero. This will lead to a zero inflow and outflow.")

    def compute(self):
        """Determine inflows and outflows and store values in the class instance."""
        self._check_needed_arrays()
        self._compute_cohorts_and_inflow()
        self._compute_outflow()

    def _compute_cohorts_and_inflow(self):
        """With given total stock and lifetime distribution,
        the method builds the stock by cohort and the inflow.
        This involves solving the lower triangular equation system A*x=b,
        where A is the survival function matrix, x is the inflow vector, and b is the stock vector.
        """
        if self.solver == "manual":
            self._compute_inflow_manual()
        elif self.solver == "lapack":
            self._compute_inflow_lapack()
        else:
            raise ValueError(f"Unknown engine: {self.solver}")

        self._stock_by_cohort = np.einsum(
            "c...,tc...->tc...", self.inflow.values, self.lifetime_model.sf
        )

    def _compute_inflow_manual(self) -> tuple[np.ndarray]:
        """With given total stock and lifetime distribution,
        the method builds the stock by cohort and the inflow,
        using a manual algorithm for solving of the equation system (see "solver" doc for details).
        """
        # Maths behind implementation:
        # Solve square linear equation system
        #   sf * inflow = stock
        # where sf is a lower triangular matrix (since year >= cohort)
        # => in every row i:
        #   sum_{j=1...i} (sf_i,j inflow_j) = stock_i
        # solve for inflow_i:
        #   inflow_i = ( stock_i - sum_{j=1...i-1}(sf_i,j * inflow_j) ) / sf_ii
        inflow_whole_period = np.zeros_like(self.inflow.values)
        for i in range(self._n_t):
            stock_i = self.stock.values[i, ...]
            sf_ij = self.lifetime_model.sf[i, :i, ...]
            inflow_j = inflow_whole_period[:i, ...]
            sf_ii = self.lifetime_model.sf[i, i, ...]

            inflow_whole_period[i, ...] = (stock_i - (sf_ij * inflow_j).sum(axis=0)) / sf_ii
        self.inflow.values[...] = self._to_annual(inflow_whole_period)

    def _compute_inflow_lapack(self) -> tuple[np.ndarray]:
        """With given total stock and lifetime distribution,
        the method builds the stock by cohort and the inflow,
        using lapack for solving of the equation system (see "engine" doc for details).
        """
        sf = self.lifetime_model.sf
        slt = (slice(None),)
        inflow_whole_period = np.zeros_like(self.inflow.values)
        for i in np.ndindex(self._shape_no_t):
            inflow_whole_period[slt + i] = solve_triangular(
                sf[2 * slt + i], self.stock.values[slt + i], lower=True
            )
        self.inflow.values[...] = self._to_annual(inflow_whole_period)



from pathlib import Path

import pytest

from scripts.run_notebooks import REPO_ROOT, iter_notebooks, run_notebook

NOTEBOOK_SOURCE_PATHS = iter_notebooks(REPO_ROOT)


@pytest.mark.parametrize(
    "notebook_source_path",
    NOTEBOOK_SOURCE_PATHS,
    ids=lambda notebook_source_path: str(notebook_source_path.relative_to(REPO_ROOT)),
)
def test_notebook_runs_with_zero_return_code(notebook_source_path: Path):
    result = run_notebook(notebook_path=notebook_source_path, repo_root=REPO_ROOT, inplace=False)

    assert result.returncode == 0, (
        f"Notebook execution returned {result.returncode}: {notebook_source_path}\n"
        f"STDOUT:\n{result.stdout}\n"
        f"STDERR:\n{result.stderr}"
    )



import numpy as np
import pandas as pd
from numpy.testing import assert_almost_equal, assert_array_almost_equal, assert_array_equal
from pydantic_core import ValidationError
import pytest
from copy import deepcopy

from flodym import FlodymArray, DimensionSet, Dimension

places = Dimension(name="place", letter="p", items=["Earth", "Sun", "Moon", "Venus"])
local_places = Dimension(name="local place", letter="l", items=["Earth"])
time = Dimension(name="time", letter="t", items=[1990, 2000, 2010])
historic_time = Dimension(name="historic time", letter="h", items=[1990, 2000])
animals = Dimension(name="animal", letter="a", items=["cat", "mouse"])

base_dim_list = [places, time]

dims = DimensionSet(dim_list=base_dim_list)
values = np.random.rand(4, 3)
numbers = FlodymArray(name="two", dims=dims, values=values)

dims_incl_animals = DimensionSet(dim_list=base_dim_list + [animals])
animal_values = np.random.rand(4, 3, 2)
space_animals = FlodymArray(name="space_animals", dims=dims_incl_animals, values=animal_values)


def test_flodym_array_validations():
    dims = DimensionSet(dim_list=[local_places, time])

    # example with values with the correct shape
    FlodymArray(name="numbers", dims=dims, values=np.array([[1, 2, 3]]))

    # example with dimensions reversed
    with pytest.raises(ValidationError):
        FlodymArray(name="numbers", dims=dims, values=np.array([[1], [2], [3]]))

    # example with too many values
    with pytest.raises(ValidationError):
        FlodymArray(name="numbers", dims=dims, values=np.array([[1, 2, 3, 4]]))

    # example with no values passed -> filled with zeros
    zero_values = FlodymArray(name="numbers", dims=dims)
    assert zero_values.values.shape == (1, 3)
    assert np.all([zero_values.values == 0])


def test_cast_to():
    # example of duplicating values along new axis (e.g. same number of cats and mice)
    casted_flodym_array = numbers.cast_to(target_dims=dims_incl_animals)
    assert casted_flodym_array.dims == dims_incl_animals
    assert casted_flodym_array.values.shape == (4, 3, 2)
    assert_almost_equal(np.sum(casted_flodym_array.values), 2 * np.sum(values))

    # example with differently ordered dimensions
    target_dims = DimensionSet(dim_list=[animals] + base_dim_list[::-1])
    casted_flodym_array = numbers.cast_to(target_dims=target_dims)
    assert casted_flodym_array.values.shape == (2, 3, 4)


def test_sum_to():
    # sum over one dimension
    summed_flodym_array = space_animals.sum_to(result_dims=("p", "t"))
    assert summed_flodym_array.dims == DimensionSet(dim_list=base_dim_list)
    assert_array_almost_equal(summed_flodym_array.values, np.sum(animal_values, axis=2))

    # sum over two dimensions
    summed_flodym_array = space_animals.sum_to(result_dims=("t"))
    assert_array_almost_equal(
        summed_flodym_array.values, np.sum(np.sum(animal_values, axis=2), axis=0)
    )

    # example attempt to get a resulting dimension that does not exist
    with pytest.raises(KeyError):
        space_animals.sum_to(result_dims=("s"))

    # example where dimensions to sum over are specified rather than the remaining dimensions
    summed_over = space_animals.sum_over(sum_over_dims=("p", "a"))
    assert_array_almost_equal(summed_over.values, summed_flodym_array.values)

    # example sum over dimension that doesn't exist
    with pytest.raises(KeyError):
        space_animals.sum_over(sum_over_dims=("s"))


def test_get_shares_over():
    # example of getting shares over one dimension
    shares = space_animals.get_shares_over(dim_letters=("p"))
    assert shares.dims == space_animals.dims
    wanted_values = np.einsum("pta,ta->pta", animal_values, 1 / np.sum(animal_values, axis=0))
    assert_array_almost_equal(shares.values, wanted_values)

    # example of getting shares over two dimensions
    shares = space_animals.get_shares_over(dim_letters=("p", "a"))
    wanted_values = np.einsum("pta,t->pta", animal_values, 1 / np.sum(animal_values, axis=(0, 2)))
    assert_array_almost_equal(shares.values, wanted_values)

    # example of getting shares over all dimensions
    shares = space_animals.get_shares_over(dim_letters=("p", "t", "a"))
    assert_array_almost_equal(shares.values, animal_values / np.sum(animal_values))

    # example of getting shares over a dimension that doesn't exist
    with pytest.raises(AssertionError):
        space_animals.get_shares_over(dim_letters=("s",))


def test_maths():
    # test minimum
    minimum = space_animals.minimum(numbers)
    assert minimum.dims == dims
    assert_array_almost_equal(minimum.values, np.minimum(values, animal_values.sum(axis=2)))

    # test maximum
    maximum = space_animals.maximum(numbers)
    assert maximum.dims == dims
    assert_array_almost_equal(maximum.values, np.maximum(values, animal_values.sum(axis=2)))

    # test sum
    summed = space_animals + numbers
    assert summed.dims == dims
    assert_array_almost_equal(summed.values, animal_values.sum(axis=2) + values)

    # test minus
    subtracted = space_animals - numbers
    assert subtracted.dims == dims
    assert_array_almost_equal(subtracted.values, animal_values.sum(axis=2) - values)
    subtracted_flipped = numbers - space_animals
    assert subtracted_flipped.dims == dims
    assert_array_almost_equal(subtracted_flipped.values, values - animal_values.sum(axis=2))

    # test multiply
    multiplied = numbers * space_animals
    assert multiplied.dims == dims_incl_animals  # different from behaviour of above methods
    assert_array_almost_equal(multiplied.values[:, :, 0], values * animal_values[:, :, 0])
    assert_array_almost_equal(multiplied.values[:, :, 1], values * animal_values[:, :, 1])

    # test divide
    divided = space_animals / numbers
    assert divided.dims == dims_incl_animals
    assert_array_almost_equal(divided.values[:, :, 0], animal_values[:, :, 0] / values)
    assert_array_almost_equal(divided.values[:, :, 1], animal_values[:, :, 1] / values)
    divided_flipped = numbers / space_animals
    assert divided_flipped.dims == dims_incl_animals
    assert_array_almost_equal(divided_flipped.values[:, :, 0], values / (animal_values[:, :, 0]))
    assert_array_almost_equal(divided_flipped.values[:, :, 1], values / (animal_values[:, :, 1]))


def test_sub_array_handler():
    space_cat = space_animals["cat"]  # space cat from str
    another_space_cat = space_animals[{"a": "cat"}]  # space cat from dict
    assert_array_equal(space_cat.values, another_space_cat.values)

    space_1990 = space_animals[{"t": 1990}]  # space animals in 1990
    assert space_1990.values.shape == (4, 2)
    assert space_1990.dims.letters == ("p", "a")

    with pytest.raises(ValueError):
        space_animals[{"a": "dog"}]  # there isn't a dog in space_animals


def test_dimension_subsets():
    historic_dims_incl_animals = DimensionSet(dim_list=[places, historic_time, animals])
    historic_space_animals = FlodymArray(dims=historic_dims_incl_animals)
    historic_space_animals[...] = space_animals[{"t": historic_time}]

    assert np.min(historic_space_animals.values) > 0.0

    space_animals_copy = FlodymArray(dims=dims_incl_animals)
    space_animals_copy[{"t": historic_time}] = 1.0 * historic_space_animals
    space_animals_copy[{"t": 2010}] = 1.0 * space_animals[{"t": 2010}]

    assert_array_equal(space_animals.values, space_animals_copy.values)

    wrong_historic_time = (Dimension(name="historic time", letter="t", items=[1990, 2000]),)
    with pytest.raises(ValueError):
        space_animals[{"t": wrong_historic_time}]  # same letter in original and subset

    another_wrong_historic_time = (Dimension(name="historic time", letter="p", items=[1990, 2000]),)
    with pytest.raises(ValueError):
        space_animals[
            {"t": another_wrong_historic_time}
        ]  # same letter in other dim of original and subset

    with pytest.raises(ValueError):
        historic_space_animals[{"h": time}]  # index is not a subset


class TestFlodymArrayIndexing:
    """Tests for getitem and setitem, including with advanced indexing separated by slices."""

    # 80 years
    years = Dimension(name="year", letter="t", items=list(range(2020, 2100)))
    # 12 regions
    regions = Dimension(name="region", letter="r", items=[f"R{i}" for i in range(1, 13)])
    # 4 sectors
    sectors = Dimension(name="sector", letter="s", items=["S1", "S2", "S3", "S4"])
    # 7 products
    products = Dimension(
        name="product", letter="p", items=["P1", "P2", "P3", "P4", "P5", "P6", "P7"]
    )
    # 2 materials
    materials = Dimension(name="material", letter="m", items=["M1", "M2"])
    full_dims = DimensionSet(dim_list=[years, regions, sectors, products, materials])
    arr = FlodymArray(
        dims=full_dims,
        values=np.arange(np.prod(full_dims.shape)).reshape(full_dims.shape).astype(float),
    )
    # 5 regions in subset
    subset_regions = Dimension(name="sub_regions", letter="x", items=["R1", "R2", "R3", "R4", "R5"])
    # 3 products in subset
    subset_products = Dimension(name="sub_products", letter="y", items=["P1", "P2", "P3"])
    # mask includes one material, subset regions and subset products
    mask = {"m": "M1", "r": subset_regions, "p": subset_products}

    def test_get_item(self):
        cats_on_the_moon = space_animals["Moon"]["cat"]
        assert isinstance(cats_on_the_moon, FlodymArray)
        assert_array_almost_equal(cats_on_the_moon.values, space_animals.values[2, :, 0])
        # note that this does not work for the time dimension (not strings)
        # and also assumes that no item appears in more than one dimension

    def test_getitem_indexing_with_slice(self):
        """Test getitem with advanced indices separated by slices (NumPy dimension reordering)."""
        result = self.arr[self.mask]
        expected_shape = (80, 5, 4, 3)
        assert (
            result.shape == expected_shape
        ), f"getitem: Expected {expected_shape}, got {result.shape}"
        expected_values = self.arr[{"m": "M1"}][{"r": self.subset_regions}][
            {"p": self.subset_products}
        ]
        assert_array_equal(result.values, expected_values.values)

    def test_setitem(self):
        array_1d = FlodymArray(dims=dims["t",])
        array_1d[...] = 1
        array_1d[1990] = 2
        assert_array_equal(array_1d.values, np.array([2, 1, 1]))

        array_2d = FlodymArray(dims=dims["p", "t"])
        array_2d[...] = 1
        array_2d[1990] = 2
        assert_array_equal(
            array_2d.values,
            np.array([[2, 1, 1]] * 4),
        )

    def test_setitem_indexing_with_slice(self):
        """Test setitem with advanced indices separated by slices (NumPy dimension reordering)."""
        arr_copy = self.arr.copy()

        slice_dims = DimensionSet(
            dim_list=[self.years, self.subset_regions, self.sectors, self.subset_products]
        )
        new_values = FlodymArray.full(slice_dims, fill_value=123.0)
        arr_copy[self.mask] = new_values
        result = arr_copy[self.mask]
        assert np.allclose(result.values, 123), "setitem failed to update values correctly"


def test_to_df():
    fda = deepcopy(space_animals)
    fda.values[...] = 0.0
    fda.values[0, 1, 0] = 1.0

    df = fda.to_df()

    assert df.shape == (24, 1)
    assert df.loc[("Earth", 2000, "cat"), "value"] == 1.0
    assert df.loc[("Earth", 2000, "mouse"), "value"] == 0.0

    df = fda.to_df(sparse=True)

    assert df.shape == (1, 1)
    assert df.loc[("Earth", 2000, "cat"), "value"] == 1.0
    with pytest.raises(KeyError):
        df.loc[("Earth", 2000, "mouse"), "value"]


def test_from_df_strips_whitespace_when_enabled():
    selected_places = Dimension(name="place", letter="p", items=["Earth", "Sun"])
    df = pd.DataFrame(
        {
            " place ": [" Earth ", " Earth ", " Sun ", " Sun "],
            " time ": [1990, 2000, 1990, 2000],
            " value ": [1.0, 2.0, 3.0, 4.0],
        }
    )
    dims_subset = DimensionSet(dim_list=[selected_places, historic_time])

    result = FlodymArray.from_df(dims=dims_subset, df=df, strip_whitespace=True)

    assert_array_equal(result.values, np.array([[1.0, 2.0], [3.0, 4.0]]))


def test_from_df_does_not_strip_whitespace_when_disabled():
    selected_places = Dimension(name="place", letter="p", items=["Earth", "Sun"])
    df = pd.DataFrame(
        {
            " place ": [" Earth ", " Earth ", " Sun ", " Sun "],
            " time ": [1990, 2000, 1990, 2000],
            " value ": [1.0, 2.0, 3.0, 4.0],
        }
    )
    dims_subset = DimensionSet(dim_list=[selected_places, historic_time])

    with pytest.raises(ValueError, match="More than one value columns"):
        FlodymArray.from_df(dims=dims_subset, df=df, strip_whitespace=False)


class TestFlodymArrayFull:
    """Tests for FlodymArray.full() class method with various fill value types."""

    def test_full_with_int(self):
        filled = FlodymArray.full(dims, fill_value=3, name="int_filled")
        assert filled.name == "int_filled"
        assert filled.dims == dims
        assert filled.dims is not dims  # ensure dims is copied
        assert np.all(filled.values == 3)
        assert filled.values.dtype == np.int_

    def test_full_with_float(self):
        filled = FlodymArray.full(dims, fill_value=3.5, name="float_filled")
        assert filled.name == "float_filled"
        assert filled.dims == dims
        assert np.all(filled.values == 3.5)
        assert filled.values.dtype == np.float64

    def test_full_with_np_array(self):
        """Test full with a numpy array that broadcasts across dimensions."""
        # Create a 1D array that broadcasts along the time dimension
        time_values = np.array([1.0, 2.0, 3.0])  # shape (3,) for 3 time steps
        filled = FlodymArray.full(dims, fill_value=time_values, name="np_array_filled")
        assert filled.name == "np_array_filled"
        assert filled.values.shape == dims.shape  # (4, 3)
        # Each row should have [1.0, 2.0, 3.0]
        assert_array_equal(filled.values[0], time_values)
        assert_array_equal(filled.values[1], time_values)
        assert_array_equal(filled.values[3], time_values)

    def test_full_with_2d_np_array(self):
        """Test full with a 2D numpy array matching the full shape."""
        fill_array = np.arange(12).reshape(4, 3).astype(np.float64)
        filled = FlodymArray.full(dims, fill_value=fill_array, name="2d_array_filled")
        assert filled.name == "2d_array_filled"
        assert_array_equal(filled.values, fill_array)
        assert filled.values.dtype == np.float64

    def test_full_dims_not_modified(self):
        """Ensure the original dims is not modified when creating a full array."""
        original_dims = DimensionSet(dim_list=base_dim_list)
        filled = FlodymArray.full(original_dims, fill_value=1.0)
        # Modify the filled array's dims
        filled.dims.dim_list[0] = animals
        # Original should be unchanged
        assert original_dims[0] == places


class TestFlodymArrayFullLike:
    """Tests for FlodymArray.full_like() class method with various fill value types."""

    def test_full_like_with_int(self):
        template = FlodymArray.full(dims, fill_value=2.0, name="template")
        filled_like = FlodymArray.full_like(template, fill_value=5, name="int_filled")
        assert filled_like.name == "int_filled"
        assert filled_like.dims == template.dims
        assert filled_like.dims is not template.dims  # ensure dims is copied
        assert np.all(filled_like.values == 5)
        assert filled_like.values.dtype == np.int_

    def test_full_like_with_float(self):
        template = FlodymArray.full(dims, fill_value=2.0, name="template")
        filled_like = FlodymArray.full_like(template, fill_value=-1.5, name="float_filled")
        assert filled_like.name == "float_filled"
        assert np.all(filled_like.values == -1.5)
        assert filled_like.values.dtype == np.float64

    def test_full_like_with_np_array(self):
        """Test full_like with a numpy array that broadcasts across dimensions."""
        template = FlodymArray.full(dims, fill_value=2.0)
        # Create a 1D array that broadcasts along the time dimension
        time_values = np.array([10.0, 20.0, 30.0])  # shape (3,) for 3 time steps
        filled_like = FlodymArray.full_like(template, fill_value=time_values)
        assert filled_like.values.shape == dims.shape  # (4, 3)
        # Each row should have [10.0, 20.0, 30.0]
        assert_array_equal(filled_like.values[0], time_values)
        assert_array_equal(filled_like.values[1], time_values)
        assert_array_equal(filled_like.values[2], time_values)

    def test_full_like_with_2d_np_array(self):
        """Test full_like with a 2D numpy array matching the full shape."""
        template = FlodymArray.full(dims, fill_value=2.0)
        fill_array = np.arange(12).reshape(4, 3).astype(np.int32)
        filled_like = FlodymArray.full_like(template, fill_value=fill_array)
        assert_array_equal(filled_like.values, fill_array)
        assert filled_like.values.dtype == np.int32

    def test_full_like_with_explicit_dtype(self):
        """Test that explicit dtype parameter overrides fill_value dtype."""
        template = FlodymArray.full(dims, fill_value=2.0)
        filled_like = FlodymArray.full_like(template, fill_value=5, dtype=np.float64)
        assert np.all(filled_like.values == 5.0)
        assert filled_like.values.dtype == np.float64

    def test_full_like_template_not_modified(self):
        """Ensure the template array is not modified."""
        template = FlodymArray.full(dims, fill_value=2.0, name="template")
        original_values = template.values.copy()
        filled_like = FlodymArray.full_like(template, fill_value=-1.0)
        # Modify filled_like
        filled_like.values[...] = 999
        filled_like.dims.dim_list[0] = animals
        # Template should be unchanged
        assert_array_equal(template.values, original_values)
        assert template.dims[0] == places


class TestFlodymArrayScalar:
    """Tests for FlodymArray.scalar() class method with various value types."""

    def test_scalar_with_int(self):
        scalar = FlodymArray.scalar(42, name="int_scalar")
        assert scalar.name == "int_scalar"
        assert len(scalar.dims) == 0
        assert scalar.dims.letters == ()
        assert scalar.shape == ()
        assert scalar.size == 1
        assert scalar.values.shape == ()
        assert scalar.values.item() == 42

    def test_scalar_with_float(self):
        scalar = FlodymArray.scalar(4.2, name="float_scalar")
        assert scalar.name == "float_scalar"
        assert scalar.values.item() == pytest.approx(4.2)

    def test_scalar_with_np_float(self):
        scalar = FlodymArray.scalar(np.float32(2.718), name="np_float_scalar")
        assert scalar.values.item() == pytest.approx(2.718, rel=1e-5)
        assert scalar.values.dtype == np.float32


class TestFlodymArrayCopy:
    """Tests for FlodymArray.copy() method ensuring deep copy behavior."""

    def test_copy_returns_new_object(self):
        array = FlodymArray.full(dims, fill_value=4, name="original")
        array_copy = array.copy()
        assert array_copy is not array

    def test_copy_dims_equal_but_not_same_object(self):
        """Ensure dims is deep copied, not just referenced."""
        array = FlodymArray.full(dims, fill_value=4, name="original")
        array_copy = array.copy()
        assert array_copy.dims == array.dims
        assert array_copy.dims is not array.dims

    def test_copy_dims_internal_dimensions_not_shared(self):
        """Ensure individual Dimension objects within dims are also copied."""
        array = FlodymArray.full(dims, fill_value=4, name="original")
        array_copy = array.copy()
        # Check that internal dimension list is not the same object
        assert array_copy.dims.dim_list is not array.dims.dim_list

    def test_copy_values_equal_but_not_same_object(self):
        """Ensure values array is deep copied, not just referenced."""
        array = FlodymArray.full(dims, fill_value=4, name="original")
        array_copy = array.copy()
        assert_array_equal(array_copy.values, array.values)
        assert array_copy.values is not array.values

    def test_copy_name_preserved(self):
        """Ensure name is copied correctly."""
        array = FlodymArray.full(dims, fill_value=4, name="my_array")
        array_copy = array.copy()
        assert array_copy.name == array.name
        assert array_copy.name == "my_array"

    def test_copy_values_modification_independent(self):
        """Modifying copied values should not affect original."""
        array = FlodymArray.full(dims, fill_value=4, name="original")
        array_copy = array.copy()
        array_copy.values[...] = 999
        assert np.all(array.values == 4)
        assert np.all(array_copy.values == 999)

    def test_copy_dims_modification_independent(self):
        """Modifying copied dims should not affect original."""
        array = FlodymArray.full(dims, fill_value=4, name="original")
        array_copy = array.copy()
        # Try to modify the copy's dims
        array_copy.dims.dim_list[0] = animals
        # Original should be unchanged
        assert array.dims[0] == places


if __name__ == "__main__":
    t = TestFlodymArrayIndexing()
    t.test_getitem_indexing_with_slice()



import pytest

from flodym import (
    DataReader,
    Dimension,
    DimensionDefinition,
    FlowDefinition,
    MFADefinition,
    MFASystem,
)


class _SingleDimDataReader(DataReader):
    """Minimal reader for tests with one dimension and no parameters."""

    def __init__(self, items_by_dimension_name: dict[str, list]):
        self.items_by_dimension_name = items_by_dimension_name

    def read_dimension(self, dimension_definition: DimensionDefinition) -> Dimension:
        return Dimension(
            name=dimension_definition.name,
            letter=dimension_definition.letter,
            items=self.items_by_dimension_name[dimension_definition.name],
            dtype=dimension_definition.dtype,
        )

    def read_parameter_values(self, parameter_name, dims):
        raise AssertionError(f"No parameters expected, but got request for {parameter_name}.")


class _MinimalMFASystem(MFASystem):
    def compute(self):
        pass


def test_mfa_system_from_definition_without_stocks():
    definition = MFADefinition(
        dimensions=[DimensionDefinition(name="time", letter="t", dtype=int)],
        processes=["sysenv"],
        flows=[],
    )
    data_reader = _SingleDimDataReader(items_by_dimension_name={"time": [2020]})

    mfa = _MinimalMFASystem.from_data_reader(definition=definition, data_reader=data_reader)

    assert mfa.stocks == {}
    assert mfa.flows == {}
    assert mfa.parameters == {}
    assert list(mfa.processes.keys()) == ["sysenv"]


def test_mass_balance_check_without_stocks_for_simple_system():
    definition = MFADefinition(
        dimensions=[DimensionDefinition(name="time", letter="t", dtype=int)],
        processes=["sysenv", "process"],
        flows=[
            FlowDefinition(
                from_process_name="sysenv",
                to_process_name="process",
                dim_letters=("t",),
            ),
            FlowDefinition(
                from_process_name="process",
                to_process_name="sysenv",
                dim_letters=("t",),
            ),
        ],
    )
    data_reader = _SingleDimDataReader(items_by_dimension_name={"time": [2020]})

    mfa = _MinimalMFASystem.from_data_reader(definition=definition, data_reader=data_reader)

    mfa.flows["sysenv => process"][...] = 2.0
    mfa.flows["process => sysenv"][...] = 2.0
    mfa.check_mass_balance(raise_error=True)

    mfa.flows["process => sysenv"][...] = 1.0
    with pytest.raises(ValueError, match="Mass balance check failed"):
        mfa.check_mass_balance(tolerance=1e-12, raise_error=True)



"""Home to helper functions for working with `FlodymArray`s."""

from .flodym_arrays import FlodymArray
from .dimensions import Dimension


def flodym_array_stack(
    flodym_arrays: list[FlodymArray], dimension: Dimension, inplace: bool = False
) -> FlodymArray:
    """Stack a list of FlodymArray objects using a new dimension.
    Like numpy.stack, but for `FlodymArray`s. New dimension is added at the end.

    Args:
        flodym_arrays (list[FlodymArray]): List of FlodymArray objects to stack. Must have the same dimensions and shape.
        dimension (Dimension): The new dimension to stack along. Must not be present in the FlodymArrays to stack.
        inplace (bool, optional): If True, modify the first FlodymArray in place. If False, return a new FlodymArray object.
    """
    if len(flodym_arrays) != dimension.len:
        raise ValueError(
            f"Length of flodym_arrays ({len(flodym_arrays)}) must match length of dimension ({dimension.len})"
        )
    for flodym_array in flodym_arrays:
        if flodym_array.dims != flodym_arrays[0].dims:
            raise ValueError("All FlodymArrays to stack must have the same dimensions")
    if dimension.letter in flodym_array.dims:
        raise ValueError(f"Dimension {dimension.letter} already present in FlodymArrays to stack")
    flodym_array0 = flodym_arrays[0]
    extended_dimensions = flodym_array0.dims.append(dimension)
    extended = flodym_array0.cast_to(extended_dimensions, inplace=inplace)
    for item, flodym_array in zip(dimension.items, flodym_arrays):
        extended[{dimension.letter: item}] = flodym_array
    return extended



import numpy as np
import pytest
import time

from flodym.dimensions import Dimension, DimensionSet
from flodym.flodym_arrays import StockArray
from flodym.mfa_definition import StockDefinition
from flodym.stock_helper import make_empty_stocks
from flodym.stocks import InflowDrivenDSM, StockDrivenDSM
from flodym.lifetime_models import LogNormalLifetime

dim_list = [
    Dimension(
        name="time",
        letter="t",
        items=list(range(1900, 2101)),
        dtype=int,
    ),
    Dimension(
        name="product",
        letter="a",
        items=["automotive", "construction"],
        dtype=str,
    ),
]

dims = DimensionSet(dim_list=dim_list)


def test_stocks():
    inflow_values = np.exp(-np.linspace(-2, 2, 201) ** 2)
    inflow_values = np.stack([inflow_values, inflow_values]).T
    inflow = StockArray(dims=dims, values=inflow_values)

    lifetime_model = LogNormalLifetime(dims=dims, time_letter="t", mean=60, std=25)
    inflow_driven_dsm = InflowDrivenDSM(
        dims=dims,
        inflow=inflow,
        lifetime_model=lifetime_model,
        time_letter="t",
    )
    inflow_driven_dsm.compute()
    stock_fda = inflow_driven_dsm.stock

    stock_driven_dsm = StockDrivenDSM(
        dims=dims,
        stock=stock_fda,
        lifetime_model=lifetime_model,
        time_letter="t",
        solver="manual",
    )
    stock_driven_dsm.compute()
    inflow_post = stock_driven_dsm.inflow
    assert np.allclose(inflow.values, inflow_post.values)

    stock_driven_dsm = StockDrivenDSM(
        dims=dims,
        stock=stock_fda,
        lifetime_model=lifetime_model,
        time_letter="t",
        solver="lapack",
    )
    stock_driven_dsm.compute()
    inflow_post_invert = stock_driven_dsm.inflow
    assert np.allclose(inflow.values, inflow_post_invert.values)
    # return inflow, inflow_post, inflow_post_invert


def test_lifetime_quadrature():
    # Put in constant inflow and check stationary stock values

    # Long lifetimes:
    # Inflow at start/end of time step under/overestimate stock by half a year,
    # others should work well
    inflow, stocks = get_stocks_by_quadrature(mean=30, std=10)
    targets = {
        "ltm_start": 29.5,
        "ltm_end": 30.5,
        "ltm_middle": 30,
        "ltm_2": 30,
        "ltm_6": 30,
    }
    eps = 0.01
    for name, stock in stocks.items():
        assert np.abs(stock["automotive"].values[-1] - targets[name]) < eps

    # Short lifetimes:
    # only high-order quadrature should work well
    inflow, stocks = get_stocks_by_quadrature(mean=0.3, std=0.1)
    for name, stock in stocks.items():
        if name == "ltm_6":
            assert np.abs(stock["automotive"].values[-1] - 0.3) < eps
        else:
            assert np.abs(stock["automotive"].values[-1] - 0.3) > eps


def get_stocks_by_quadrature(mean, std):
    inflow_values = np.exp(-np.linspace(-2, 2, 201) ** 2)
    inflow_values = np.stack([inflow_values, inflow_values]).T
    inflow_values = np.ones_like(inflow_values)
    inflow = StockArray(dims=dims, values=inflow_values)

    lifetime_models = {
        "ltm_start": LogNormalLifetime(
            dims=dims, time_letter="t", mean=mean, std=std, inflow_at="start"
        ),
        "ltm_end": LogNormalLifetime(
            dims=dims, time_letter="t", mean=mean, std=std, inflow_at="end"
        ),
        "ltm_middle": LogNormalLifetime(
            dims=dims, time_letter="t", mean=mean, std=std, inflow_at="middle"
        ),
        "ltm_2": LogNormalLifetime(
            dims=dims, time_letter="t", mean=mean, std=std, n_pts_per_interval=2
        ),
        "ltm_6": LogNormalLifetime(
            dims=dims, time_letter="t", mean=mean, std=std, n_pts_per_interval=6
        ),
    }
    stocks = {}
    for name, lifetime_model in lifetime_models.items():
        inflow_driven_dsm = InflowDrivenDSM(
            dims=dims,
            inflow=inflow,
            lifetime_model=lifetime_model,
            time_letter="t",
        )
        inflow_driven_dsm.compute()
        stocks[name] = inflow_driven_dsm.stock
    return inflow, stocks


def test_unequal_time_steps():
    """Unequal time step lengths should not influence the development of the stock apart from numerical errors."""
    # case with all years
    t_all = Dimension(
        name="time",
        letter="t",
        items=list(range(1900, 2101)),
        dtype=int,
    )
    # case with only a few select unequally spaced years
    t_select_items = [i for i in t_all.items if i % 13 == 0 or i % 7 == 0]
    t_select = Dimension(
        name="time",
        letter="s",
        items=t_select_items,
        dtype=int,
    )
    stocks = []
    for t in t_all, t_select:
        dims = DimensionSet(dim_list=[t])
        lifetime_model = LogNormalLifetime(
            dims=dims, time_letter=t.letter, mean=5, std=2, n_pts_per_interval=6
        )
        inflow_values = np.ones(t.len)
        inflow = StockArray(dims=dims, values=inflow_values)
        inflow_driven_dsm = InflowDrivenDSM(
            dims=dims,
            inflow=inflow,
            lifetime_model=lifetime_model,
            time_letter=t.letter,
        )
        inflow_driven_dsm.compute()
        stocks.append(inflow_driven_dsm.stock)
    stocks_all = stocks[0][{"t": t_select}]
    values_all = stocks_all.values[-10:]
    stocks_select = stocks[1]
    values_select = stocks_select.values[-10:]
    assert np.max(np.abs(values_all - values_select)) < 0.01


def test_make_empty_stocks_raises_clear_error_for_missing_nondefault_time_letter():
    time_dim = Dimension(name="time", letter="s", items=[2000, 2005, 2010], dtype=int)
    product_dim = Dimension(name="product", letter="a", items=["automotive"], dtype=str)
    stock_dims = DimensionSet(dim_list=[time_dim, product_dim])
    stock_definition = StockDefinition(
        name="stock_with_nondefault_time",
        dim_letters=("s", "a"),
        subclass=InflowDrivenDSM,
        lifetime_model_class=LogNormalLifetime,
    )

    with pytest.raises(ValueError, match="time_letter"):
        make_empty_stocks([stock_definition], processes={}, dims=stock_dims)


def test_make_empty_stocks_accepts_explicit_nondefault_time_letter():
    time_dim = Dimension(name="time", letter="s", items=[2000, 2005, 2010], dtype=int)
    product_dim = Dimension(name="product", letter="a", items=["automotive"], dtype=str)
    stock_dims = DimensionSet(dim_list=[time_dim, product_dim])
    stock_definition = StockDefinition(
        name="stock_with_nondefault_time",
        dim_letters=("s", "a"),
        time_letter="s",
        subclass=InflowDrivenDSM,
        lifetime_model_class=LogNormalLifetime,
    )

    stocks = make_empty_stocks([stock_definition], processes={}, dims=stock_dims)

    assert stocks["stock_with_nondefault_time"].time_letter == "s"



from pydantic_core import ValidationError
import pytest

from flodym import Dimension, DimensionSet


def test_validate_dimension_set():
    # example valid DimensionSet
    dimensions = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
    ]
    DimensionSet(dim_list=dimensions)

    # example with repeated dimension letters in DimensionSet
    dimensions.append({"name": "another_time", "letter": "t", "items": [2020, 2030]})
    with pytest.raises(ValidationError) as error_msg:
        DimensionSet(dim_list=dimensions)
    assert "letter" in str(error_msg.value)


def test_get_subset():
    subset_dimensions = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
    ]
    material_dimension = {"name": "material", "letter": "m", "items": ["material_0", "material_1"]}

    parent_dimensions = subset_dimensions + [material_dimension]
    dimension_set = DimensionSet(dim_list=parent_dimensions)

    # example of subsetting the dimension set using dimension letters
    subset_from_letters = dimension_set.get_subset(dims=("t", "p"))
    assert subset_from_letters == DimensionSet(dim_list=subset_dimensions)

    # example of subsetting the dimension set using dimension names
    subset_from_names = dimension_set.get_subset(dims=("time", "place"))
    assert subset_from_names == subset_from_letters

    # example where the requested subset dimension doesn't exist
    with pytest.raises(KeyError):
        dimension_set.get_subset(dims=("s", "p"))


def test_index_with_letters_and_names():
    """Test that index() method accepts both letters and names."""
    dimensions = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
        {"name": "material", "letter": "m", "items": ["steel", "aluminum"]},
    ]
    dimension_set = DimensionSet(dim_list=dimensions)

    # Test with letters
    assert dimension_set.index("t") == 0
    assert dimension_set.index("p") == 1
    assert dimension_set.index("m") == 2

    # Test with names
    assert dimension_set.index("time") == 0
    assert dimension_set.index("place") == 1
    assert dimension_set.index("material") == 2

    # Test with non-existent key
    with pytest.raises(KeyError):
        dimension_set.index("nonexistent")


def test_size_with_letters_and_names():
    """Test that size() method accepts both letters and names."""
    dimensions = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
    ]
    dimension_set = DimensionSet(dim_list=dimensions)

    # Test with letters
    assert dimension_set.size("t") == 3
    assert dimension_set.size("p") == 1

    # Test with names
    assert dimension_set.size("time") == 3
    assert dimension_set.size("place") == 1


def test_drop_with_letters_and_names():
    """Test that drop() method accepts both letters and names."""
    dimensions = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
        {"name": "material", "letter": "m", "items": ["steel", "aluminum"]},
    ]
    dimension_set = DimensionSet(dim_list=dimensions)

    # Test drop with letter
    dropped = dimension_set.drop("t")
    assert dropped.letters == ("p", "m")
    assert dropped.names == ("place", "material")

    # Test drop with name
    dropped = dimension_set.drop("place")
    assert dropped.letters == ("t", "m")
    assert dropped.names == ("time", "material")


def test_replace_with_letters_and_names():
    """Test that replace() method accepts both letters and names."""
    dimensions = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
    ]
    dimension_set = DimensionSet(dim_list=dimensions)

    new_dim = Dimension(name="NewDim", letter="n", items=[1, 2, 3])

    # Test replace with letter
    replaced = dimension_set.replace("t", new_dim)
    assert replaced.names == ("NewDim", "place")
    assert replaced.letters == ("n", "p")

    # Test replace with name
    replaced = dimension_set.replace("place", new_dim)
    assert replaced.names == ("time", "NewDim")
    assert replaced.letters == ("t", "n")


def test_intersection_operator():
    """Test the intersection operator (&) for DimensionSets."""
    dims1 = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
        {"name": "material", "letter": "m", "items": ["steel", "aluminum"]},
    ]
    dims2 = [
        {"name": "place", "letter": "p", "items": ["World"]},
        {"name": "material", "letter": "m", "items": ["steel", "aluminum"]},
        {"name": "product", "letter": "r", "items": ["car", "bike"]},
    ]
    set1 = DimensionSet(dim_list=dims1)
    set2 = DimensionSet(dim_list=dims2)

    # Test intersection using & operator
    intersection = set1 & set2
    assert intersection.letters == ("p", "m")
    assert intersection.names == ("place", "material")

    # Test that it's equivalent to intersect_with method
    assert intersection == set1.intersect_with(set2)

    # Test intersection with no common dimensions
    dims3 = [{"name": "animal", "letter": "a", "items": ["cat", "dog"]}]
    set3 = DimensionSet(dim_list=dims3)
    empty_intersection = set1 & set3
    assert len(empty_intersection.dim_list) == 0

    # Test intersection with same set
    same_intersection = set1 & set1
    assert same_intersection == set1


def test_union_operator():
    """Test the union operator (|) for DimensionSets."""
    dims1 = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
    ]
    dims2 = [
        {"name": "place", "letter": "p", "items": ["World"]},
        {"name": "material", "letter": "m", "items": ["steel", "aluminum"]},
    ]
    set1 = DimensionSet(dim_list=dims1)
    set2 = DimensionSet(dim_list=dims2)

    # Test union using | operator
    union = set1 | set2
    assert union.letters == ("t", "p", "m")
    assert union.names == ("time", "place", "material")

    # Test that it's equivalent to union_with method
    assert union == set1.union_with(set2)

    # Test union with disjoint sets
    dims3 = [{"name": "animal", "letter": "a", "items": ["cat", "dog"]}]
    set3 = DimensionSet(dim_list=dims3)
    disjoint_union = set1 | set3
    assert disjoint_union.letters == ("t", "p", "a")
    assert disjoint_union.names == ("time", "place", "animal")

    # Test union with same set
    same_union = set1 | set1
    assert same_union == set1


def test_difference_operator():
    """Test the difference operator (-) for DimensionSets."""
    dims1 = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
        {"name": "material", "letter": "m", "items": ["steel", "aluminum"]},
    ]
    dims2 = [
        {"name": "place", "letter": "p", "items": ["World"]},
        {"name": "product", "letter": "r", "items": ["car", "bike"]},
    ]
    set1 = DimensionSet(dim_list=dims1)
    set2 = DimensionSet(dim_list=dims2)

    # Test difference using - operator
    difference = set1 - set2
    assert difference.letters == ("t", "m")
    assert difference.names == ("time", "material")

    # Test that it's equivalent to difference_with method
    assert difference == set1.difference_with(set2)

    # Test difference with disjoint sets
    dims3 = [{"name": "animal", "letter": "a", "items": ["cat", "dog"]}]
    set3 = DimensionSet(dim_list=dims3)
    disjoint_difference = set1 - set3
    assert disjoint_difference == set1

    # Test difference with same set
    same_difference = set1 - set1
    assert len(same_difference.dim_list) == 0


def test_symmetric_difference_operator():
    """Test the symmetric difference operator (^) for DimensionSets."""
    dims1 = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
        {"name": "material", "letter": "m", "items": ["steel", "aluminum"]},
    ]
    dims2 = [
        {"name": "place", "letter": "p", "items": ["World"]},
        {"name": "product", "letter": "r", "items": ["car", "bike"]},
    ]
    set1 = DimensionSet(dim_list=dims1)
    set2 = DimensionSet(dim_list=dims2)

    # Test symmetric difference using ^ operator
    sym_diff = set1 ^ set2
    assert set(sym_diff.letters) == {"t", "m", "r"}
    assert set(sym_diff.names) == {"time", "material", "product"}

    # Test that it's equivalent to (set1 - set2) | (set2 - set1)
    expected = (set1 - set2) | (set2 - set1)
    assert sym_diff.letters == expected.letters

    # Test symmetric difference with same set
    same_sym_diff = set1 ^ set1
    assert len(same_sym_diff.dim_list) == 0

    # Test symmetric difference with disjoint sets
    dims3 = [{"name": "animal", "letter": "a", "items": ["cat", "dog"]}]
    set3 = DimensionSet(dim_list=dims3)
    disjoint_sym_diff = set1 ^ set3
    # Should contain all dimensions from both sets
    assert set(disjoint_sym_diff.letters) == {"t", "p", "m", "a"}


def test_len_dunder():
    """Test the __len__ dunder method for DimensionSets."""
    # Empty dimension set
    empty_set = DimensionSet(dim_list=[])
    assert len(empty_set) == 0

    # Single dimension
    dims1 = [{"name": "time", "letter": "t", "items": [1990, 2000, 2010]}]
    set1 = DimensionSet(dim_list=dims1)
    assert len(set1) == 1

    # Multiple dimensions
    dims2 = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
        {"name": "material", "letter": "m", "items": ["steel", "aluminum"]},
    ]
    set2 = DimensionSet(dim_list=dims2)
    assert len(set2) == 3

    # Test that len is equivalent to ndim property
    assert len(set2) == set2.ndim


def test_bool_dunder():
    """Test the __bool__ dunder method for DimensionSets."""
    # Empty dimension set should be falsy
    empty_set = DimensionSet(dim_list=[])
    assert not empty_set
    assert bool(empty_set) is False

    # Non-empty dimension set should be truthy
    dims = [{"name": "time", "letter": "t", "items": [1990, 2000, 2010]}]
    non_empty_set = DimensionSet(dim_list=dims)
    assert non_empty_set
    assert bool(non_empty_set) is True

    # Test with multiple dimensions
    dims2 = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
    ]
    set2 = DimensionSet(dim_list=dims2)
    assert set2
    assert bool(set2) is True


def test_ndim_property():
    """Test that the ndim property still works correctly after being moved."""
    # Empty dimension set
    empty_set = DimensionSet(dim_list=[])
    assert empty_set.ndim == 0

    # Single dimension
    dims1 = [{"name": "time", "letter": "t", "items": [1990, 2000, 2010]}]
    set1 = DimensionSet(dim_list=dims1)
    assert set1.ndim == 1

    # Multiple dimensions
    dims2 = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
        {"name": "material", "letter": "m", "items": ["steel", "aluminum"]},
    ]
    set2 = DimensionSet(dim_list=dims2)
    assert set2.ndim == 3


def test_dimensionset_empty():
    """DimensionSet.empty should yield a reusable scalar-friendly container."""
    empty = DimensionSet.empty()
    assert isinstance(empty, DimensionSet)
    assert len(empty) == 0
    assert empty.shape == ()
    assert empty.letters == ()
    assert DimensionSet.empty() is not empty


def test_dimensionset_total_size():
    """total_size should equal the product of dimension lengths."""
    dims = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["Earth", "Moon"]},
    ]
    dim_set = DimensionSet(dim_list=dims)
    assert dim_set.shape == (3, 2)
    assert dim_set.total_size == 6


def test_dimension_as_dimset():
    """Test that as_dimset() converts a Dimension to a DimensionSet."""
    dim = Dimension(name="time", letter="t", items=[1990, 2000, 2010])

    # Convert to DimensionSet
    dimset = dim.as_dimset()

    # Verify it's a DimensionSet
    assert isinstance(dimset, DimensionSet)

    # Verify it contains only the one dimension
    assert len(dimset) == 1
    assert dimset.letters == ("t",)
    assert dimset.names == ("time",)

    # Verify the dimension is the same
    assert dimset[0] == dim


def test_dimension_add():
    """Test the __add__ operator for Dimension objects."""
    dim1 = Dimension(name="time", letter="t", items=[1990, 2000, 2010])
    dim2 = Dimension(name="place", letter="p", items=["World", "Moon"])

    # Test Dimension + Dimension
    dimset = dim1 + dim2
    assert isinstance(dimset, DimensionSet)
    assert len(dimset) == 2
    assert dimset.letters == ("t", "p")
    assert dimset.names == ("time", "place")

    # Test Dimension + DimensionSet
    dim3 = Dimension(name="material", letter="m", items=["steel", "aluminum"])
    dimset2 = dim1 + dim2
    dimset3 = dim3 + dimset2

    assert isinstance(dimset3, DimensionSet)
    assert len(dimset3) == 3
    assert dimset3.letters == ("m", "t", "p")
    assert dimset3.names == ("material", "time", "place")

    # Test adding invalid type
    with pytest.raises(TypeError):
        dim1 + "invalid"


def test_dimensionset_append():
    """Test the append() method for DimensionSet."""
    dims = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
    ]
    dimset = DimensionSet(dim_list=dims)
    new_dim = Dimension(name="material", letter="m", items=["steel", "aluminum"])

    # Test append with inplace=False (returns new DimensionSet)
    new_dimset = dimset.append(new_dim, inplace=False)
    assert new_dimset is not None
    assert len(new_dimset) == 3
    assert new_dimset.letters == ("t", "p", "m")

    # Original should be unchanged
    assert len(dimset) == 2
    assert dimset.letters == ("t", "p")

    # Test append with inplace=True (modifies in place)
    dimset2 = DimensionSet(dim_list=dims)
    result = dimset2.append(new_dim, inplace=True)
    assert result is None
    assert len(dimset2) == 3
    assert dimset2.letters == ("t", "p", "m")

    # Test appending dimension with duplicate letter
    dup_dim = Dimension(name="other_time", letter="t", items=[2020, 2030])
    with pytest.raises(ValueError):
        dimset.append(dup_dim)

    # Test appending non-Dimension object
    with pytest.raises(TypeError):
        dimset.append("invalid")


def test_dimensionset_prepend():
    """Test the prepend() method for DimensionSet."""
    dims = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
    ]
    dimset = DimensionSet(dim_list=dims)
    new_dim = Dimension(name="material", letter="m", items=["steel", "aluminum"])

    # Test prepend with inplace=False (returns new DimensionSet)
    new_dimset = dimset.prepend(new_dim, inplace=False)
    assert new_dimset is not None
    assert len(new_dimset) == 3
    assert new_dimset.letters == ("m", "t", "p")

    # Original should be unchanged
    assert len(dimset) == 2
    assert dimset.letters == ("t", "p")

    # Test prepend with inplace=True (modifies in place)
    dimset2 = DimensionSet(dim_list=dims)
    result = dimset2.prepend(new_dim, inplace=True)
    assert result is None
    assert len(dimset2) == 3
    assert dimset2.letters == ("m", "t", "p")

    # Test prepending dimension with duplicate letter
    dup_dim = Dimension(name="other_time", letter="t", items=[2020, 2030])
    with pytest.raises(ValueError):
        dimset.prepend(dup_dim)

    # Test prepending non-Dimension object
    with pytest.raises(TypeError):
        dimset.prepend("invalid")


def test_dimensionset_remove():
    """Test the remove() method (alias for drop()) for DimensionSet."""
    dims = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
        {"name": "material", "letter": "m", "items": ["steel", "aluminum"]},
    ]
    dimset = DimensionSet(dim_list=dims)

    # Test remove with inplace=False (returns new DimensionSet)
    new_dimset = dimset.remove("m", inplace=False)
    assert new_dimset is not None
    assert len(new_dimset) == 2
    assert new_dimset.letters == ("t", "p")

    # Original should be unchanged
    assert len(dimset) == 3
    assert dimset.letters == ("t", "p", "m")

    # Test remove with inplace=True (modifies in place)
    dimset2 = DimensionSet(dim_list=dims)
    result = dimset2.remove("m", inplace=True)
    assert result is None
    assert len(dimset2) == 2
    assert dimset2.letters == ("t", "p")

    # Test removing with name instead of letter
    dimset3 = DimensionSet(dim_list=dims)
    new_dimset3 = dimset3.remove("time", inplace=False)
    assert len(new_dimset3) == 2
    assert new_dimset3.letters == ("p", "m")


def test_dimensionset_insert():
    """Test the insert() method for DimensionSet."""
    dims = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
    ]
    dimset = DimensionSet(dim_list=dims)
    new_dim = Dimension(name="material", letter="m", items=["steel", "aluminum"])

    # Test insert at beginning (index 0)
    new_dimset = dimset.insert(0, new_dim, inplace=False)
    assert new_dimset is not None
    assert len(new_dimset) == 3
    assert new_dimset.letters == ("m", "t", "p")

    # Test insert at middle (index 1)
    new_dimset2 = dimset.insert(1, new_dim, inplace=False)
    assert len(new_dimset2) == 3
    assert new_dimset2.letters == ("t", "m", "p")

    # Test insert at end (index 2)
    new_dimset3 = dimset.insert(2, new_dim, inplace=False)
    assert len(new_dimset3) == 3
    assert new_dimset3.letters == ("t", "p", "m")

    # Original should be unchanged
    assert len(dimset) == 2
    assert dimset.letters == ("t", "p")

    # Test insert with inplace=True
    dimset2 = DimensionSet(dim_list=dims)
    result = dimset2.insert(1, new_dim, inplace=True)
    assert result is None
    assert len(dimset2) == 3
    assert dimset2.letters == ("t", "m", "p")

    # Test inserting dimension with duplicate letter
    dup_dim = Dimension(name="other_time", letter="t", items=[2020, 2030])
    with pytest.raises(ValueError):
        dimset.insert(0, dup_dim)

    # Test inserting non-Dimension object
    with pytest.raises(TypeError):
        dimset.insert(0, "invalid")


def test_dimensionset_add():
    """Test the __add__ operator for DimensionSet."""
    dims1 = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
    ]
    dims2 = [
        {"name": "material", "letter": "m", "items": ["steel", "aluminum"]},
        {"name": "product", "letter": "r", "items": ["car", "bike"]},
    ]
    dimset1 = DimensionSet(dim_list=dims1)
    dimset2 = DimensionSet(dim_list=dims2)

    # Test DimensionSet + DimensionSet with no overlap
    result = dimset1 + dimset2
    assert isinstance(result, DimensionSet)
    assert len(result) == 4
    assert result.letters == ("t", "p", "m", "r")

    # Test DimensionSet + Dimension
    new_dim = Dimension(name="animal", letter="a", items=["cat", "dog"])
    result2 = dimset1 + new_dim
    assert isinstance(result2, DimensionSet)
    assert len(result2) == 3
    assert result2.letters == ("t", "p", "a")

    # Test that adding overlapping dimensions raises ValueError
    dims3 = [
        {"name": "place", "letter": "p", "items": ["Mars"]},  # 'p' overlaps
        {"name": "product", "letter": "r", "items": ["car", "bike"]},
    ]
    dimset3 = DimensionSet(dim_list=dims3)

    with pytest.raises(ValueError) as error_msg:
        dimset1 + dimset3
    assert "overlap" in str(error_msg.value).lower()

    # Test commutativity for disjoint sets
    result_forward = dimset1 + dimset2
    result_backward = dimset2 + dimset1
    assert set(result_forward.letters) == set(result_backward.letters)


def test_copy_dimension_set():
    """Test the copy method of DimensionSet."""
    dimensions = [
        Dimension(name="time", letter="t", items=[1990, 2000, 2010]),
        Dimension(name="place", letter="p", items=["World"]),
    ]
    original = DimensionSet(dim_list=dimensions)

    # Create a copy
    copied = original.copy()

    # Ensure the copied object is not the same as the original
    assert copied is not original

    # Ensure modifying the copy does not affect the original
    assert copied.dim_list is not original.dim_list

    # Ensure the contained Dimension objects remain the same
    for o_dim, c_dim in zip(original, copied):
        assert o_dim is c_dim



import pytest
import numpy as np

from flodym import Dimension, DimensionSet, FlodymArray
from flodym.export.array_plotter import PyplotArrayPlotter, PlotlyArrayPlotter


@pytest.fixture
def test_array():
    """Create a test array with multiple dimensions."""
    dims = DimensionSet(
        dim_list=[
            Dimension(name="Time", letter="t", items=[2020, 2021, 2022]),
            Dimension(name="Region", letter="r", items=["EU", "US"]),
        ]
    )
    values = np.random.rand(3, 2)
    return FlodymArray(dims=dims, values=values, name="TestArray")


def test_pyplot_plotter_with_names(test_array):
    """Test PyplotArrayPlotter with dimension names."""
    plotter = PyplotArrayPlotter(array=test_array, intra_line_dim="Time", linecolor_dim="Region")
    assert plotter.intra_line_dim == "Time"
    assert plotter.linecolor_dim == "Region"
    plotter.plot()  # Ensure plotting works without error


def test_pyplot_plotter_with_letters(test_array):
    """Test PyplotArrayPlotter with dimension letters."""
    plotter = PyplotArrayPlotter(array=test_array, intra_line_dim="t", linecolor_dim="r")
    assert plotter.intra_line_dim == "t"
    assert plotter.linecolor_dim == "r"
    plotter.plot()  # Ensure plotting works without error


def test_pyplot_plotter_with_mixed(test_array):
    """Test PyplotArrayPlotter with mixed dimension names and letters."""
    plotter = PyplotArrayPlotter(array=test_array, intra_line_dim="Time", linecolor_dim="r")
    assert plotter.intra_line_dim == "Time"
    assert plotter.linecolor_dim == "r"
    plotter.plot()  # Ensure plotting works without error


def test_plotly_plotter_with_names(test_array):
    """Test PlotlyArrayPlotter with dimension names."""
    plotter = PlotlyArrayPlotter(array=test_array, intra_line_dim="Time", linecolor_dim="Region")
    assert plotter.intra_line_dim == "Time"
    assert plotter.linecolor_dim == "Region"
    plotter.plot()  # Ensure plotting works without error


def test_plotly_plotter_with_letters(test_array):
    """Test PlotlyArrayPlotter with dimension letters."""
    plotter = PlotlyArrayPlotter(array=test_array, intra_line_dim="t", linecolor_dim="r")
    assert plotter.intra_line_dim == "t"
    assert plotter.linecolor_dim == "r"
    plotter.plot()  # Ensure plotting works without error


def test_plotly_plotter_with_mixed(test_array):
    """Test PlotlyArrayPlotter with mixed dimension names and letters."""
    plotter = PlotlyArrayPlotter(array=test_array, intra_line_dim="Time", linecolor_dim="r")
    assert plotter.intra_line_dim == "Time"
    assert plotter.linecolor_dim == "r"
    plotter.plot()  # Ensure plotting works without error


def test_plotter_with_subplot_dim_letter(test_array):
    """Test that subplot_dim also accepts letters."""
    # Create array with 3 dimensions
    dims = DimensionSet(
        dim_list=[
            Dimension(name="Time", letter="t", items=[2020, 2021]),
            Dimension(name="Region", letter="r", items=["EU", "US"]),
            Dimension(name="Element", letter="e", items=["C", "Fe"]),
        ]
    )
    values = np.random.rand(2, 2, 2)
    array = FlodymArray(dims=dims, values=values, name="TestArray3D")

    plotter = PyplotArrayPlotter(
        array=array, intra_line_dim="t", linecolor_dim="r", subplot_dim="e"
    )
    assert plotter.subplot_dim == "e"
    plotter.plot()  # Ensure plotting works without error


def test_plotter_invalid_dimension_name(test_array):
    """Test that invalid dimension names/letters raise an error."""
    with pytest.raises(ValueError, match="not in array dimensions"):
        PyplotArrayPlotter(array=test_array, intra_line_dim="invalid", linecolor_dim="r")


def test_plotter_invalid_dimension_letter(test_array):
    """Test that invalid dimension letters raise an error."""
    with pytest.raises(ValueError, match="not in array dimensions"):
        PyplotArrayPlotter(array=test_array, intra_line_dim="t", linecolor_dim="x")
